use crate::myrand::Rand;
use crate::utxo_store::{Height, UtxoInfo, UtxoStore};
use bitcoin::absolute::LockTime;
use bitcoin::key::Secp256k1;
use bitcoin::secp256k1::{PublicKey, SecretKey};
use bitcoin::transaction::Version;
use bitcoin::{
    Address, Amount, CompressedPublicKey, OutPoint, ScriptBuf, Sequence, Transaction, TxIn, TxOut,
    Witness,
};

/// Represents a block, with list of transactions. First tx is the coinbase.
#[derive(Clone)]
pub struct Block {
    pub height: Height,
    pub txs: Vec<Transaction>,
}

impl Block {
    pub fn new(height: Height, txs: Vec<Transaction>) -> Self {
        Self { height, txs }
    }

    pub fn tx_count(&self) -> usize {
        self.txs.len()
    }

    pub fn print_summary(&self) {
        print!("block at h {}  w/ {} txs, ", self.height, self.txs.len())
    }
}

pub trait SimulatorSink {
    fn connect_block(&mut self, block: &Block);
}

/// Simulate a chain from the beginning to the given height, for testing chain processors, UTXO stores (e.g. UTreeXO)
/// Note that to be able to generate realistic-looking valid transactions, this simulator
/// also includes a primitive UTXO store.
pub struct ChainSimulator {
    rng: Rand,
    sink: Box<dyn SimulatorSink>,
    utxos: UtxoStore,
    blocks: Vec<Block>,
}

impl ChainSimulator {
    pub fn new(sink: Box<dyn SimulatorSink>, rng_seed: u64) -> Self {
        Self {
            rng: Rand::new_from_seed(rng_seed),
            sink,
            utxos: UtxoStore::new(),
            blocks: Vec::new(),
        }
    }

    fn random_address(&mut self) -> Address {
        let mut b32 = <[u8; 32]>::default();
        for i in 0..32 {
            b32[i] = self.rng.random_byte();
        }
        let random_sk = SecretKey::from_slice(&b32).unwrap();
        let secp = Secp256k1::new();
        let random_pubkey = CompressedPublicKey(PublicKey::from_secret_key(&secp, &random_sk));
        Address::p2shwpkh(&random_pubkey, bitcoin::Network::Bitcoin)
    }

    fn random_coinbase(&mut self) -> Transaction {
        Transaction {
            version: Version::ONE,
            lock_time: bitcoin::absolute::LockTime::ZERO,
            input: Vec::new(), // coinbase
            output: vec![TxOut {
                script_pubkey: self.random_address().script_pubkey(),
                value: Amount::ZERO, // dummy
            }],
        }
    }

    /// Add the given block to the chain
    fn process_block(&mut self, block: Block) {
        self.blocks.push(block);
    }

    /// Add the tx to the Utxo store, remove inputs
    fn process_tx(&mut self, tx: &Transaction, height: Height) {
        let txid = tx.compute_txid();
        for oidx in 0..tx.output.len() {
            self.utxos.add(UtxoInfo::new(
                OutPoint {
                    txid,
                    vout: oidx as u32,
                },
                height,
            ))
        }
    }

    pub fn run(&mut self, end_height: Height) {
        let mut height = 0;
        while height <= end_height {
            self.generate_block(height);
            height += 1;
        }
    }

    fn generate_block(&mut self, height: Height) {
        let txs = self.generate_transactions(height);
        let block = Block::new(height, txs);
        self.process_block(block.clone());
        self.sink.connect_block(&block);
    }

    fn simul_tx_count_per_block(&mut self, height: Height) -> usize {
        if height <= 3 {
            return 0;
        }
        if height <= 20 {
            return 1;
        }
        if height < 100 {
            return self.rng.random_range(10) as usize;
        }
        if height < 1000 {
            return self.rng.random_range(25) as usize;
        }
        if height < 50000 {
            return self.rng.random_range(100) as usize;
        }
        self.rng.random_range(4000) as usize
    }

    fn simul_tx_input_count(&mut self) -> usize {
        1 + self.rng.random_range(2) as usize
    }

    fn simul_tx_output_count(&mut self) -> usize {
        1 + self.rng.random_range(2) as usize
    }

    fn generate_transaction(&mut self, current_height: Height) -> Transaction {
        let ic = self.simul_tx_input_count();
        let oc = self.simul_tx_output_count();
        let mut input = Vec::<TxIn>::new();
        for _i in 0..ic {
            let previous_output = self.utxos.get_random(&mut self.rng, current_height);
            let txin = TxIn {
                previous_output: previous_output.clone(),
                script_sig: ScriptBuf::new(),
                sequence: Sequence::MAX,
                witness: Witness::new(),
            };

            // Remove input UTXO right away
            if !self.utxos.remove(&previous_output) {
                panic!(
                    "Processing a tx with unknown input, op {:?}",
                    previous_output
                );
            }

            input.push(txin);
        }
        let mut output = Vec::<TxOut>::new();
        for _o in 0..oc {
            let txo = TxOut {
                value: Amount::ZERO,
                script_pubkey: self.random_address().script_pubkey(),
            };
            output.push(txo.clone());
        }
        let tx = Transaction {
            lock_time: LockTime::ZERO,
            version: Version::ONE,
            input,
            output,
        };
        tx
    }

    fn generate_transactions(&mut self, height: Height) -> Vec<Transaction> {
        let mut txs = Vec::<Transaction>::new();
        let coinbase = self.random_coinbase();
        self.process_tx(&coinbase, height);
        txs.push(coinbase);

        let tx_count = self.simul_tx_count_per_block(height);
        for _t in 0..tx_count {
            let tx = self.generate_transaction(height);
            self.process_tx(&tx, height);
            txs.push(tx);
        }

        txs
    }

    pub fn tip_height(&self) -> Height {
        (self.blocks.len() - 1) as Height
    }

    pub fn print(&self, verbose: bool) {
        println!(
            "Chain with {} blocks  (tip {}),  UTXO count: {}",
            self.blocks.len(),
            self.tip_height(),
            self.utxos.count(),
        );
        if verbose {
            for b in &self.blocks {
                b.print_summary();
                println!();
            }
        }
    }
}

#[cfg(test)]
mod tests {
    use super::{Block, ChainSimulator, SimulatorSink};

    struct TestSink {}

    impl SimulatorSink for TestSink {
        fn connect_block(&mut self, block: &Block) {
            println!(
                "Block connected, height {}, tx_cnt {}",
                block.height,
                block.tx_count()
            )
        }
    }

    #[test]
    fn test_chain_simulator_100() {
        let sink = Box::new(TestSink {});

        let mut simul = ChainSimulator::new(sink, 1);
        simul.run(100);

        assert_eq!(simul.tip_height(), 100);
    }
}
