use crate::chain_simulator::{Block, SimulatorSink};
use crate::memtree_wrapper::MemTreeWrapper;
use bitcoin::hashes::Hash;
use rustreexo::accumulator::node_hash::BitcoinNodeHash as NodeHash;

/// Tester for MemTree full accumulator
pub struct MemTreeSink {
    acc: MemTreeWrapper,
}

impl MemTreeSink {
    pub fn new() -> Self {
        Self {
            acc: MemTreeWrapper::new(),
        }
    }
}

impl SimulatorSink for MemTreeSink {
    fn connect_block(&mut self, block: &Block) {
        println!(
            "MemTree: connect_block, height {}, tx_cnt {}",
            block.height,
            block.tx_count()
        );

        // Process transactions one-by-one
        // This is conceptually not correct, but this is how the issue can be reproduced
        for tx in &block.txs {
            let mut inphash = Vec::<NodeHash>::new();
            let mut outhash = Vec::<NodeHash>::new();
            let txid = tx.compute_txid();
            for inp in &tx.input {
                let hash = hash_for_out_point(
                    &inp.previous_output.txid.as_byte_array(),
                    inp.previous_output.vout,
                );
                inphash.push(NodeHash::Some(hash));

                // verify each input one-by-one
                if let Err(err) = self.acc.prove(&[NodeHash::Some(hash)]) {
                    panic!("Input not found!!! {}", err);
                }
            }
            for vout in 0..tx.output.len() {
                let hash = hash_for_out_point(&txid.as_byte_array(), vout as u32);
                outhash.push(NodeHash::Some(hash));
            }

            if let Err(err) = self.acc.modify(&outhash, &inphash) {
                panic!("Acc error!!! {}", err);
            }
        }
    }
}

fn hash_for_out_point(txid: &[u8; 32], vout: u32) -> [u8; 32] {
    let mut hash = txid.clone();
    hash[31] = hash[31] ^ ((vout & 0x000000FF) as u8);
    hash[30] = hash[30] ^ (((vout & 0x0000FF00) >> 8) as u8);
    hash
}

#[cfg(test)]
mod tests {
    use super::MemTreeSink;
    use crate::chain_simulator::ChainSimulator;

    fn memtree_tester_n(n: u32) {
        let sink = Box::new(MemTreeSink::new());

        let mut simul = ChainSimulator::new(sink, 1);
        simul.run(n);

        assert_eq!(simul.tip_height(), n);
    }

    #[test]
    fn test_rustreexo_memtree_tester_20() {
        memtree_tester_n(20);
    }

    #[test]
    fn test_rustreexo_memtree_tester_100() {
        memtree_tester_n(100);
    }

    #[test]
    fn test_rustreexo_memtree_tester_2000() {
        memtree_tester_n(2000);
    }
}
