use crate::myrand::Rand;
use bitcoin::OutPoint;
use std::collections::HashMap;

pub type Height = u32;

#[derive(Clone, Debug, PartialEq)]
pub struct UtxoInfo {
    txo: OutPoint,
    added: Height,
    is_removed: bool,
}

impl UtxoInfo {
    pub fn new(txo: OutPoint, added: Height) -> Self {
        Self {
            txo,
            added,
            is_removed: false,
        }
    }
}

/// Hold the current set of UTXOs
pub struct UtxoStore {
    /// The elements. The ones deleted stay here, with is_removed set
    store: Vec<UtxoInfo>,
    /// Map by OutPoint, map to index in store
    map: HashMap<OutPoint, usize>,
}

impl UtxoStore {
    pub fn new() -> Self {
        Self {
            store: Vec::new(),
            map: HashMap::new(),
        }
    }

    pub fn add(&mut self, u: UtxoInfo) {
        let idx = self.store.len();
        self.store.push(u.clone());
        let key = u.txo;
        self.map.insert(key, idx);
    }

    pub fn count(&self) -> usize {
        self.map.len()
    }

    #[cfg(test)]
    pub fn get(&self, txo: &OutPoint) -> Option<&UtxoInfo> {
        match self.map.get(txo) {
            None => None,
            Some(idx) => Some(&self.store[*idx]),
        }
    }

    pub fn remove(&mut self, txo: &OutPoint) -> bool {
        match self.map.get(txo) {
            None => panic!("Attempt to remove non-existent OP {:?}", txo),
            Some(idx) => {
                self.store[*idx].is_removed = true;
                self.map.remove(txo);
                true
            }
        }
    }

    pub fn get_random(&self, rng: &mut Rand, current_height: Height) -> OutPoint {
        let mut try_count = 0;
        loop {
            let idx = rng.random_range(self.store.len() as u64) as usize;
            let utxo = &self.store[idx];
            if !utxo.is_removed {
                if utxo.added < current_height {
                    return utxo.txo.clone();
                }
            }
            try_count += 1;
            if try_count > self.store.len() * 10 {
                panic!(
                    "Tried too many times, {}, len {}",
                    try_count,
                    self.store.len()
                );
            }
        }
    }
}

#[cfg(test)]
mod test {
    use bitcoin::{hashes::Hash, OutPoint, Txid};

    use super::{UtxoInfo, UtxoStore};

    #[test]
    fn test_utxo_store() {
        let mut st = UtxoStore::new();
        let txo1 = OutPoint::new(Txid::from_slice(&[1; 32]).unwrap(), 2);
        let utxo1 = UtxoInfo::new(txo1.clone(), 20);
        st.add(utxo1.clone());

        assert_eq!(st.count(), 1);
        assert_eq!(st.get(&txo1).unwrap(), &utxo1);

        let txo2 = OutPoint::new(Txid::from_slice(&[2; 32]).unwrap(), 1);
        let utxo2 = UtxoInfo::new(txo2.clone(), 40);
        st.add(utxo2.clone());

        assert_eq!(st.count(), 2);
        assert_eq!(st.get(&txo1).unwrap(), &utxo1);
        assert_eq!(st.get(&txo2).unwrap(), &utxo2);

        st.remove(&txo1);

        assert_eq!(st.count(), 1);
        assert!(st.get(&txo1).is_none());
        assert_eq!(st.get(&txo2).unwrap(), &utxo2);
    }
}
